# -*- coding: UTF-8 -*-

import glob
import os
import json
import tempfile


class FileQueueManager:
    """ Класс менеджера очередей на файловой системе """
    def __init__(self, queueRoot=tempfile.gettempdir(), chunkSize=25000, log=None):
        self.queueRoot = queueRoot
        self.chunkSize = chunkSize
        self.log = log

        # создание корневой директории очередей
        # здесь можно свалиться, но это нормально
        if not os.path.exists(self.queueRoot):
            os.makedirs(self.queueRoot)

    def getShopDests(self, prefixQueueName):
        """ Получение списка наименований магазинов на основании имён очередей
        Args:
            prefixQueueName(str): префикс очереди
        Returns:
            list: список наименований магазинов
        """
        dirs = glob.glob(os.path.join(self.queueRoot, '%s*' % prefixQueueName))
        return map(lambda x: x.split('.')[-1], dirs)

    def removeUncompletedTransactions(self, queueName, source):
        """ Удаление неполных транзакций. Транзакция считается неполной, если у неё transactionId > 0
            и отсутствует сообщение о завершении транзакции.
            Этот метод должен вызывать тот модуль, который создавал транзакцию, указав соответствующий source
        Args:
            queueName(str): имя очереди
            source(str): создатель транзакции(sync-core2, maint, artix-pos, exchangers-upload)
        """
        # 1. получение всех транзакций
        files = glob.glob(os.path.join(self.queueRoot, queueName, '*.objs'))
        transactions = set(map(lambda x: int(os.path.splitext(os.path.basename(x))[0]) >> 32, files))

        # 2. получение множества незавершенных транзакций
        uncompleted = filter(lambda x: not self._isTransactionComplete(queueName, [source], x), transactions)

        # 3. удаление незавершенных транзакций
        for transactionId in uncompleted:
            self.removeTransaction(queueName, transactionId)

    def beginTransaction(self, queueName):
        """ Начало транзакции. Возвращает идентификатор транзакции
        Args:
            queueName(str): имя очереди
        Returns:
            int: идентификатор транзакции
        """
        return self._getNextTransactionId(os.path.join(self.queueRoot, queueName))

    def endTransaction(self, queueName, transactionId, message):
        """ Конец транзакции. Создание сообщения с метаданными транзакции
        Args:
            queueName(str): имя очереди
            transactionId(int): идентификатор транзакции
            message(json): данные сообщения
        """
        queuePath = os.path.join(self.queueRoot, queueName)
        if not os.path.exists(queuePath):
            os.makedirs(queuePath)
        with open(os.path.join(queuePath, self._getNextFileName(queuePath, transactionId)), 'w') as trFile:
            json.dump(message, trFile)

    def enqueue(self, queueName, transactionId, message):
        """ Помещение сообщения в очередь. При этом сообщение будет в отдельном файле
        Args:
            queueName(str): имя очереди
            transactionId(int): номер транзакции
            message(dict): сообщение
        """
        queuePath = os.path.join(self.queueRoot, queueName)
        if not os.path.exists(queuePath):
            os.makedirs(queuePath)
        nextFileName = self._getNextFileName(queuePath, transactionId)
        fileName = os.path.join(queuePath, nextFileName)
        with open(fileName, 'w') as outfile:
            json.dump(message, outfile)

    def getCompleteTransactionId(self, queueName):
        """ Получение идентификатора первой полной тразнакции
        Args:
            queueName(str): имя очереди
        Returns:
            int: идентификатор транзакции или None, если транзакций нет
        """
        files = glob.glob(os.path.join(self.queueRoot, queueName, '*.objs'))
        allTransactions = sorted(set(map(lambda x: int(os.path.splitext(os.path.basename(x))[0]) >> 32, files)))
        completedTransactions = sorted(filter(lambda x: self._isTransactionComplete(queueName,
                ['sync-core2', 'maint', 'artix-pos', 'exchangers-upload'], x), allTransactions))
        if completedTransactions:
            return completedTransactions[0]
        return None

    def getNextMessage(self, queueName, transactionId):
        """ Получение следующего сообщения транзакции
        Args:
            queueName(str): имя очереди
            transactionId(int): идентификатор транзакции
        Returns:
            dict: следующее сообщение транзакции
        """
        files = glob.glob(os.path.join(self.queueRoot, queueName, '*.objs'))
        transactionFiles = filter(lambda x: (int(os.path.splitext(os.path.basename(x))[0]) >> 32) == transactionId, files)
        for messagePath in sorted(transactionFiles, key=lambda x: int(os.path.splitext(os.path.basename(x))[0])):
            try:
                with open(messagePath, 'r') as messageFile:
                    yield json.load(messageFile)
            except Exception as e:
                if self.log:
                    self.log.warn('Ошибка чтения файла %s. %s' % (messagePath, str(e)))

    def getTransactionProperties(self, queueName, transactionId):
        if self.log:
            self.log.info('Получение свойств транзакции %s из очереди %s' % (transactionId, queueName))
        files = glob.glob(os.path.join(self.queueRoot, queueName, '*.objs'))
        transactionFiles = filter(lambda x: (int(os.path.splitext(os.path.basename(x))[0]) >> 32) == transactionId, files)
        if not transactionFiles:
            return None
        lastFileName = max(transactionFiles, key=lambda x: int(os.path.splitext(os.path.basename(x))[0]))
        with open(lastFileName, 'r') as lastFile:
            data = json.load(lastFile)
            return data.get('properties')

    def dequeue(self, messagePath):
        """ Удаление сообщения по его полному пути
        Args:
            messagePath(str): полный путь до сообщения
        """
        if self.log:
            self.log.debug('Удаление сообщения "%s"' % messagePath)
        if messagePath and os.path.exists(messagePath):
            os.remove(messagePath)

    def removeTransaction(self, queueName, transactionId):
        """ Удаление всех файлов транзакции по идентификатору транзакции
        Args:
            queueName(str): имя очереди
            transactionId(int): идентификатор транзакции
        """
        if self.log:
            self.log.info('Удаление транзакции с идентификатором %s' % transactionId)
        files = glob.glob(os.path.join(self.queueRoot, queueName, '*.objs'))
        transactionFiles = filter(lambda x: (int(os.path.splitext(os.path.basename(x))[0]) >> 32) == transactionId, files)
        for trFile in transactionFiles:
            try:
                os.remove(trFile)
            except:
                if self.log:
                    self.log.warn('Ошибка удаления файла %s' % trFile)

    def _getNextFileName(self, queuePath, transactionId):
        """ Получение имени следующего файла данных
        Args:
            queuePath(str): полный путь до очереди
            transactionId(int): номер транзакции
        Returns:
            str: имя следующего файла данных
        """
        files = glob.glob(os.path.join(queuePath, '*.objs'))
        transactionFiles = filter(lambda x: (int(os.path.splitext(os.path.basename(x))[0]) >> 32) == transactionId, files)
        nextNumber = max(map(lambda x: int(os.path.splitext(os.path.basename(x))[0]), transactionFiles) + [transactionId << 32]) + 1
        return '%d.objs' % nextNumber

    def _getNextTransactionId(self, queuePath):
        """ Получение номера следующей транзакции
        Args:
            queuePath(str): полный путь до очереди
        Returns:
            int: номер следующей транзакции
        """
        files = glob.glob(os.path.join(queuePath, '*.objs'))
        nextNumber = (max(map(lambda x: int(os.path.splitext(os.path.basename(x))[0]), files) + [0]) >> 32) + 1
        return nextNumber

    def _isTransactionComplete(self, queueName, source, transactionId):
        """ Проверка того, что транзакция является полной
        Args:
            queueName(str): имя очереди
            source(list): список создателей транзакции(sync-core2, maint, artix-pos, exchangers-upload)
            transactionId(int): идентификатор транзакции
        Returns:
            bool: True - транзакция полна, False - иначе
        """
        files = glob.glob(os.path.join(self.queueRoot, queueName, '*.objs'))
        transactionFiles = filter(lambda x: (int(os.path.splitext(os.path.basename(x))[0]) >> 32) == transactionId, files)
        if not transactionFiles:
            return True
        maxFile = max(transactionFiles, key=lambda x: int(os.path.splitext(os.path.basename(x))[0]))
        if not os.path.exists(maxFile):
            return False
        with open(maxFile, 'r') as mf:
            try:
                data = json.load(mf)
                if data['properties']['source'] not in source or data['properties']['control'] == 'end_tran':
                    return True
            except Exception:
                pass
        return False
